/*
  Play WORDLE with an ESP32...
*/
#define CHECK true // false if you don't want to check your input is in the dict
#define NUMBER 5 // Number of letters (don't change unless you can change the dictionary)

#include "FS.h"
#include "SPIFFS.h"
#define FORMAT_SPIFFS_IF_FAILED true
//const char wordFile[] = "/Dict-12478.txt";
const char wordFile[] = "/Dict-5757.txt";

char Wordle[NUMBER + 1];
char Word[NUMBER + 1];
char Found[NUMBER + 1];
byte Alphabet[27] = {2};
byte check[NUMBER + 1] = {0}; // 2 well placed, 1 misplaced, 0 not in the word
char val[4] = "-=+";

void splashscreen() {
  Serial.println("\nWORDLE on ESP32 (by Lesept)");
  Serial.println("Guess the WORDLE in 6 tries. Each guess must be a valid 5 letter word.");
  Serial.println("My answer is:");
  Serial.printf("   %c : letter is well positionned\n", val[0]);
  Serial.printf("   %c : letter exists but not here\n", val[1]);
  Serial.printf("   %c : letter is not in the WORDLE\n", val[2]);
  Serial.println("Good luck!\n");
  Serial.printf("Please enter a %d letter word.\n", NUMBER);
  delay(200);
}

void sofar() {
  // Summarize what we know so far...
  Serial.println("Here is what you know so far...");
  Serial.printf("Well placed: %s\n", Found);
  Serial.print("Not in the Wordle: ");
  for (byte i = 0; i < 26; ++i)
    if (Alphabet[i] == 1) Serial.print(char('A' + i));
  Serial.println();
}

bool readWord() {
  while (Serial.available() == 0);
  byte n = 0;
  while (n < NUMBER) {
    if (Serial.available() > 0) {
      char letter = Serial.read();
      if (letter == '?') { // Recap what we know
        sofar();
        return false;
      }
      if (((letter >= 'a') && (letter <= 'z')) || ((letter >= 'A') && (letter <= 'Z'))) {
        if ((letter >= 'a') && (letter <= 'z')) letter = letter + 'A' - 'a';
        Word[n++] = letter;
      }
    }
  }
  Serial.println(Word);
  char t = Serial.read();
  while (t != '\n') t = Serial.read(); // Empty the buffer
  return true;
}

bool checkWord() {
  bool checkle[NUMBER + 1];
  byte nok = 0;

  // Check if letter is well placed
  for (byte i = 0; i < NUMBER; ++i) {
    checkle[i] = false;
    if (Wordle[i] == Word[i]) {
      check[i] = 2;
      Found[i] = Word[i];
      checkle[i] = true;
      ++ nok;
    } else check[i] = 0;
  }
  if (nok == NUMBER) return true; // All letters match!

  // Check if letter is in the word
  for (byte i = 0; i < NUMBER; ++i) {
    if (check[i] != 2) {
      for (byte j = 0; j < NUMBER; ++j) {
        if ((Word[i] == Wordle[j]) && !checkle[j] && check[j] != 2) {
          check[i] = 1;
          checkle[j] = true;
          break;
        }
      }
      bool isInWordle = false;
      for (byte j = 0; j < NUMBER; ++j)
        if (Word[i] == Wordle[j]) isInWordle = true;
      Alphabet[Word[i] - 'A'] = (isInWordle) ? 0 : 1;
    }
  }
  return false;
}

void pickRandom(File file) {
  // Pick a random word
  int pos = random(file.size() / (NUMBER + 1));
  file.seek(6 * pos, SeekSet);
  for (int i = 0; i < NUMBER; i++) {
    char letter = file.read();
    if ((letter >= 'a') && (letter <= 'z')) letter = letter + 'A' - 'a';
    Wordle[i] = letter;
  }
  file.close();
}


void setup ()
{
  Serial.begin (115200);
  delay(200);
  if (!SPIFFS.begin(FORMAT_SPIFFS_IF_FAILED)) {
    Serial.println("SPIFFS Mount Failed");
    return;
  }

  // Open the dictionary file
  File file = SPIFFS.open(wordFile);
  if (!file || file.isDirectory()) {
    Serial.printf("%s: Failed to open file for reading\n", wordFile);
    //  while (1);
  }
  splashscreen();
  for (byte i = 0; i < NUMBER; ++i) Found[i] = '-';
  pickRandom(file);
  //  Serial.println(Wordle); // The solution is here !
}

// Check if the proposed word is in the dictionary
bool exists() {
  char Test[NUMBER + 1];
  File file = SPIFFS.open(wordFile); // open dictionary
  int lines = file.size() / (NUMBER + 1);
  for (int i = 0; i < lines; ++i) {
    byte nok = 0;
    file.seek(6 * i, SeekSet);
    char letter = file.read(); // read first letter of each word
    if ((letter >= 'a') && (letter <= 'z')) letter = letter + 'A' - 'a';
    if (letter == Word[0]) { // if first letter matches
      nok = 1;;
      for (byte j = 1; j < NUMBER; ++j) { // read other letters
        letter = file.read();
        if ((letter >= 'a') && (letter <= 'z')) letter = letter + 'A' - 'a';
        if (letter == Word[j]) ++nok;
      }
      if (nok == NUMBER) {
        file.close();
        return true;
      }
    }
  }
  file.close();
  return false;
}

void loop() {
  static byte trials = 0; // only 6 trials
  if (readWord()) {// read the proposed word
    bool checkInput = true;
    if (CHECK) checkInput = exists(); // 450ms max
    if (!checkInput) { // test if this word is in the dictionary
      Serial.println("This word is not in my dictionary, try again");
    } else {
      if (checkWord()) { // compute the score of the word
        Serial.printf("+++++\nCongratulations, you found it in %d trials!!", trials + 1);
        while (1);
      } else {
        for (byte i = 0; i < NUMBER; ++i) Serial.print(val[check[i]]);
        Serial.println();
        if (++trials == NUMBER + 1) {
          Serial.printf("THE END !!!\nThe solution was: %s\n", Wordle);
          while (1);
        }
      }
    }
  }
}
